from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .components import UUID, CrudComponent, JsonType, KongEntity, KongError

if TYPE_CHECKING:
    from .client import Kong


class Plugin(KongEntity):
    pass


class Plugins(CrudComponent[Plugin]):
    async def create(self, **params: Any) -> Plugin:
        params = await self.preprocess_parameters(params)
        return await super().create(**params)

    async def apply_json(self, data: JsonType, clear: bool = True) -> list:
        if not isinstance(data, list):
            data = [data]
        plugins = await self.get_full_list()
        if not self.is_entity:
            plugins = [p for p in plugins if self.root_plugin(p)]
        plugin_map = {p["name"]: p for p in plugins}
        result = []
        for entry in data:
            entry = entry.copy()
            name = entry.pop("name", None)
            if not name:
                raise KongError("Plugin name not specified")
            if name in plugin_map:
                plugin = plugin_map.pop(name)
                plugin = await self.update(plugin.id, name=name, **entry)
            else:
                plugin = await self.create(name=name, **entry)

            result.append(plugin.data)
        # left over plugins
        if clear:
            for entry in plugin_map.values():
                await self.delete(entry["id"])
        return result

    def root_plugin(self, plugin: KongEntity) -> bool:
        return not (
            plugin.get("service") or plugin.get("route") or plugin.get("consumer")
        )

    async def preprocess_parameters(self, params: dict) -> dict:
        await anonymous(self.cli, params)
        if preprocessor := PLUGIN_PREPROCESSORS.get(params.get("name", "")):
            params = await preprocessor(self.cli, params)
        return params

    async def update(self, id_: str | UUID, **params: Any) -> Plugin:
        params = await self.preprocess_parameters(params)
        return await super().update(id_, **params)


class KongEntityWithPlugins(KongEntity):
    @property
    def plugins(self) -> Plugins:
        return Plugins(self, Plugin)


async def consumer_id_from_username(cli: Kong, params: dict) -> dict:
    if "id" in (params.get("consumer") or {}):
        c = await cli.consumers.get(params["consumer"]["id"])
        params["consumer"]["id"] = c["id"]
    return params


async def anonymous(cli: Kong, params: dict) -> dict:
    if "config" in params and "anonymous" in params["config"]:
        c = await cli.consumers.get(params["config"]["anonymous"])
        params["config"]["anonymous"] = c["id"]
    return params


PLUGIN_PREPROCESSORS = {"request-termination": consumer_id_from_username}



import asyncio
import json
from typing import Any, cast

import click
import yaml as _yaml
from rich.console import Console
from rich.table import Table

from . import __version__
from .client import Kong, KongError, default_admin_url
from .components import KongEntity
from .consumers import Consumer
from .utils import local_ip

admin_url = click.option(
    "--url", default=default_admin_url(), help="Kong Admin URL", show_default=True
)

as_json = click.option("--json", default=False, is_flag=True, help="Output as JSON")


@click.group()
@click.version_option(version=__version__)
def kong() -> None:
    """Kong CLI - Manage Kong Gateway configurations."""
    pass


@kong.command()
@click.argument(
    "yaml",
    type=click.File("r"),
)
@click.option(
    "--clear",
    default=False,
    is_flag=True,
    help="Clear objects not in configuration",
)
@admin_url
def yaml(
    yaml: click.File,
    clear: bool,
    url: str,
) -> None:
    "Upload a configuration from a yaml file"
    asyncio.run(_yml(yaml, clear, url))


@kong.command()
def ip() -> None:
    "Show local IP address"
    click.echo(local_ip())


@kong.command()
@click.option("-d", "--delete", type=str, help="Delete a service by name")
@admin_url
@as_json
def services(delete: str, url: str, json: bool) -> None:
    "Display services"
    asyncio.run(_services(url, as_json=json, delete=delete))


@kong.command()
@click.argument("service")
@admin_url
@as_json
def routes(service: str, url: str, json: bool) -> None:
    "Display routes for a service"
    asyncio.run(_routes(service, url, as_json=json))


@kong.command()
@admin_url
@as_json
def consumers(url: str, json: bool) -> None:
    "Display consumers"
    asyncio.run(_consumers(url, as_json=json))


@kong.command()
@click.argument("consumer")
@admin_url
def key_auth(consumer: str, url: str) -> None:
    "Create or display an authentication key for a consumer"
    asyncio.run(_auth_key(consumer, url))


async def _yml(yaml: Any, clear: bool, url: str) -> None:
    async with Kong(url=url) as cli:
        try:
            result = await cli.apply_json(_yaml.safe_load(yaml), clear=clear)
            display_json(result)
        except KongError as exc:
            raise click.ClickException(str(exc)) from None


async def _services(url: str, as_json: bool = False, delete: str | None = None) -> None:
    async with Kong(url=url) as cli:
        try:
            if delete:
                await cli.services.delete(delete)
                click.echo(f"Service '{delete}' deleted.")
                return
            services = await cli.services.get_full_list()
        except KongError as exc:
            raise click.ClickException(str(exc)) from None
        if as_json:
            display_json(services)
            return
        table = Table(title="Services")
        columns = ["Name", "Host", "Port", "Protocol", "Path", "Tags", "ID"]
        for column in columns:
            table.add_column(column)
        for s in sorted(services, key=lambda s: s.name):
            table.add_row(*[str_value(s.data, column) for column in columns])
        console = Console()
        console.print(table)


async def _routes(service: str, url: str, as_json: bool = False) -> None:
    async with Kong(url=url) as cli:
        try:
            svc = await cli.services.get(service)
            routes = await svc.routes.get_full_list()
        except KongError as exc:
            raise click.ClickException(str(exc)) from None
        if as_json:
            display_json(routes)
            return
        table = Table(title="Services")
        columns = [
            "Name",
            "Hosts",
            "Protocols",
            "Path",
            "Tags",
            "Strip Path",
            "Preserve Host",
            "ID",
        ]
        for column in columns:
            table.add_column(column)
        for s in sorted(routes, key=lambda s: s.name):
            table.add_row(*[str_value(s.data, column) for column in columns])
        console = Console()
        console.print(table)


async def _consumers(url: str, as_json: bool = False) -> None:
    async with Kong(url=url) as cli:
        try:
            consumers = await cli.consumers.get_full_list()
        except KongError as exc:
            raise click.ClickException(str(exc)) from None
        if as_json:
            display_json(consumers)
            return
        table = Table(title="Consumers")
        columns = ["Username", "ID", "Custom ID", "Tags"]
        for column in columns:
            table.add_column(column)
        for c in sorted(consumers, key=lambda s: s.name):
            table.add_row(*[str_value(c.data, column) for column in columns])
        console = Console()
        console.print(table)


async def _auth_key(consumer: str, admin_url: str) -> None:
    async with Kong(url=admin_url) as cli:
        try:
            c = cast(Consumer, await cli.consumers.get(consumer))
            keys = await c.keyauths.get_list()
            if keys:
                key = keys[0]
            else:
                key = await c.keyauths.create()
            display_json(key)
        except KongError as exc:
            raise click.ClickException(str(exc)) from None


def str_value(data: dict, key: str) -> str:
    key = key.lower().replace(" ", "_")
    value = data.get(key)
    if value is None:
        return ""
    elif isinstance(value, bool):
        return ":white_check_mark:" if value else "[red]:x:"
    elif isinstance(value, list):
        return ", ".join(str(v) for v in value)
    else:
        return str(value)


def display_json(data: Any) -> None:
    """Display data as JSON."""
    click.echo(json.dumps(data, indent=2, default=kong_entity))


def kong_entity(obj: Any) -> Any:
    if isinstance(obj, KongEntity):
        return obj.data
    raise TypeError(
        f"Cannot serialize {type(obj).__name__} to JSON"
    )  # pragma: no cover



from typing import cast

from .components import UUID, CrudComponent, JsonType
from .plugins import KongEntityWithPlugins
from .utils import as_list


class Route(KongEntityWithPlugins):
    pass


class Routes(CrudComponent):
    """Kong Routes

    Routes are always associated with a Service
    """

    async def delete(self, id_: str | UUID) -> bool:
        route = cast(KongEntityWithPlugins, self.wrap({"id": id_}))
        await route.plugins.delete_all()
        return await super().delete(id_)

    async def apply_json(self, data: JsonType, clear: bool = True) -> list[dict]:
        if not isinstance(data, list):
            data = [data]
        routes = await self.get_list()
        route_map = {r.name: r for r in routes}
        result = []
        for entry in data:
            name = entry.get("name")
            route = route_map.pop(name, None) if name else None
            entry = entry.copy()
            plugins = entry.pop("plugins", [])
            as_list("hosts", entry)
            as_list("paths", entry)
            as_list("methods", entry)
            if route:
                await self.delete(route.id)
            entity = await self.create(**entry)
            route = cast(KongEntityWithPlugins, entity)
            route.data["plugins"] = await route.plugins.apply_json(plugins)
            result.append(route.data)
        if clear:
            for route in route_map.values():
                await self.delete(route.id)
        return result



import socket
from typing import Any
from uuid import UUID

from multidict import MultiDict


def as_list(key: str, data: dict) -> dict:
    if key in data:
        v = data[key]
        if isinstance(v, str):
            v = [v]
        data[key] = list(v)
    return data


def as_dict(data: Any, key: str = "data") -> dict:
    return {key: data} if not isinstance(data, dict) else data


def local_ip() -> str:
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    try:
        return s.getsockname()[0]
    finally:
        s.close()


def as_params(*, params: dict | None = None, **kwargs: Any) -> MultiDict:
    mp: MultiDict = MultiDict(params if params is not None else {})
    mp.update(kwargs)
    return mp


def uid(id_: str | UUID) -> str:
    if isinstance(id_, UUID):
        return str(id_)
    try:
        return str(UUID(id_))
    except ValueError:
        return str(id_)



from __future__ import annotations

import os
import sys
from typing import Any, Callable, Dict, Optional

from aiohttp import ClientResponse, ClientSession

from . import __version__
from .acls import Acl, Acls
from .certificates import Certificate, Certificates
from .components import CrudComponent, KongError, KongResponseError
from .consumers import Consumer, Consumers
from .plugins import Plugin, Plugins
from .routes import Route, Routes
from .services import Service, Services
from .snis import Sni, Snis

__all__ = ["Kong", "KongError", "KongResponseError"]

DEFAULT_USER_AGENT = (
    f"python/{'.'.join(map(str, sys.version_info[:2]))} aio-kong/{__version__}"
)

KONG_ADMIN_SSL = os.getenv("KONG_ADMIN_SSL", "true").strip().lower() in (
    "true",
    "1",
    "yes",
)


def default_admin_url() -> str:
    """Return the default Kong admin URL."""
    return os.getenv("KONG_ADMIN_URL", os.getenv("KONG_URL", "http://127.0.0.1:8001"))


def kong_request_kwargs(request_kwargs: dict | None = None) -> dict:
    kwargs = request_kwargs or {}
    if "ssl" not in kwargs and not KONG_ADMIN_SSL:
        kwargs["ssl"] = False
    return kwargs


class Kong:
    """Kong client"""

    def __init__(
        self,
        url: str | None = None,
        session: ClientSession | None = None,
        request_kwargs: dict | None = None,
        user_agent: str = DEFAULT_USER_AGENT,
        content_type: str = "application/json, text/*; q=0.5",
    ) -> None:
        self.url = url or default_admin_url()
        self.session = session
        self.user_agent = user_agent
        self.content_type = content_type
        self.request_kwargs = kong_request_kwargs(request_kwargs)
        self.services = Services(self, Service)
        self.routes = Routes(self, Route)
        self.plugins = Plugins(self, Plugin)
        self.consumers = Consumers(self, Consumer)
        self.certificates = Certificates(self, Certificate)
        self.acls = Acls(self, Acl)
        self.snis = Snis(self, Sni)

    def __repr__(self) -> str:
        return self.url

    __str__ = __repr__

    @property
    def cli(self) -> Kong:
        return self

    async def close(self) -> None:
        if self.session:
            await self.session.close()

    async def __aenter__(self) -> Kong:
        return self

    async def __aexit__(self, exc_type: type, exc_val: Any, exc_tb: Any) -> None:
        await self.close()

    async def execute(
        self,
        url: str,
        method: str = "",
        headers: dict[str, str] | None = None,
        callback: Optional[Callable[[ClientResponse], Any]] = None,
        wrap: Optional[Callable[[Any], Any]] = None,
        **kw: Any,
    ) -> Any:
        if not self.session:
            self.session = ClientSession()
        method = method or "GET"
        headers_ = self.default_headers()
        headers_.update(headers or ())
        kw.update(self.request_kwargs)
        response = await self.session.request(method, url, headers=headers_, **kw)
        if callback:
            return await callback(response)
        if response.status == 204:
            return True
        if response.status >= 400:
            try:
                data = await response.json()
            except Exception:
                data = await response.text()
            raise KongResponseError(response, data)
        response.raise_for_status()
        data = await response.json()
        return wrap(data) if wrap else data

    async def apply_json(self, config: dict, clear: bool = True) -> dict:
        if not isinstance(config, dict):
            raise KongError("Expected a dict got %s" % type(config).__name__)
        result = {}
        for name, data in config.items():
            if not isinstance(data, list):
                data = [data]
            o = getattr(self, name)
            if not isinstance(o, CrudComponent):
                raise KongError("Kong object %s not available" % name)
            result[name] = await o.apply_json(data, clear=clear)
        return result

    async def delete_all(self) -> None:
        await self.services.delete_all()
        await self.consumers.delete_all()
        await self.plugins.delete_all()
        await self.certificates.delete_all()

    def default_headers(self) -> Dict[str, str]:
        return {"user-agent": self.user_agent, "accept": self.content_type}



from .components import CrudComponent, KongEntity
from .snis import Sni, Snis


class Certificate(KongEntity):
    @property
    def snis(self) -> Snis:
        return Snis(self, Sni)


class Certificates(CrudComponent[Certificate]):
    """Kong TLS certificate component"""
